/***************************************************************************
 *   Copyright (C) 1998-2009 by David Bucciarelli (davibu@interfree.it)    *
 *                                                                         *
 *   This file is part of SmallLuxGPU.                                     *
 *                                                                         *
 *   SmallLuxGPU is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   SmallLuxGPU is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#include <cstring>

#include "renderthread.h"
#include "raybuffer.h"
#include "samplebuffer.h"
#include "path.h"


RenderThread::RenderThread(unsigned int index, Scene *scn) {
	threadIndex = index;
	scene = scn;
	started = false;
}

RenderThread::~RenderThread() {
}

//------------------------------------------------------------------------------
// NativeRenderThread
//------------------------------------------------------------------------------

NativeRenderThread::NativeRenderThread(unsigned int index, NativeIntersectionDevice *device,
		Scene *scn, const bool lowLatency) : RenderThread(index, scn) {
	intersectionDevice = device;

	// Allocate buffers

	// Sample buffer
	const size_t sampleBufferSize = lowLatency ? (SAMPLE_BUFFER_SIZE / 4) : SAMPLE_BUFFER_SIZE;
	sampleBuffer = new SampleBuffer(sampleBufferSize);

	// Ray buffer (small buffers work well with CPU)
	const size_t rayBufferSize = 1024;
	sampler = new RandomSampler(lowLatency, threadIndex + 1,
		scene->camera->film->GetWidth(), scene->camera->film->GetHeight());

	pathIntegrator = new PathIntegrator(scene, sampler, sampleBuffer);
	rayBuffer = new RayBuffer(rayBufferSize);

	renderThread = NULL;
}

NativeRenderThread::~NativeRenderThread() {
	if (started)
		Stop();

	delete rayBuffer;
	delete pathIntegrator;
	delete sampler;
	delete sampleBuffer;
}

void NativeRenderThread::Start() {
	RenderThread::Start();

	sampler->Init(scene->camera->film->GetWidth(), scene->camera->film->GetHeight());
	sampleBuffer->Reset();
	rayBuffer->Reset();
	pathIntegrator->ReInit();

	// Create the thread for the rendering
	renderThread = new boost::thread(boost::bind(NativeRenderThread::RenderThreadImpl, this));
}

void NativeRenderThread::Interrupt() {
	if (renderThread)
		renderThread->interrupt();
}

void NativeRenderThread::Stop() {
	if (renderThread) {
		renderThread->interrupt();
		renderThread->join();
		delete renderThread;
		renderThread = NULL;
	}

	RenderThread::Stop();
}

void NativeRenderThread::ClearPaths() {
	pathIntegrator->ClearPaths();
}

void NativeRenderThread::RenderThreadImpl(NativeRenderThread *renderThread) {
	cerr << "[NativeRenderThread::" << renderThread->threadIndex << "] Rendering thread started" << endl;

	try {
		RayBuffer *rayBuffer = renderThread->rayBuffer;
		PathIntegrator *pathIntegrator = renderThread->pathIntegrator;
		NativeIntersectionDevice *intersectionDevice = renderThread->intersectionDevice;

		while (!boost::this_thread::interruption_requested()) {
			rayBuffer->Reset();
			pathIntegrator->FillRayBuffer(rayBuffer);
			intersectionDevice->TraceRays(rayBuffer);
			pathIntegrator->AdvancePaths(rayBuffer);
		}

		cerr << "[NativeRenderThread::" << renderThread->threadIndex << "] Rendering thread halted" << endl;
	} catch (boost::thread_interrupted) {
		cerr << "[NativeRenderThread::" << renderThread->threadIndex << "] Rendering thread halted" << endl;
	} catch (cl::Error err) {
		cerr << "[NativeRenderThread::" << renderThread->threadIndex << "] RenderingERROR: " << err.what() << "(" << err.err() << ")" << endl;
	}
}

//------------------------------------------------------------------------------
// DeviceRenderThread
//------------------------------------------------------------------------------

DeviceRenderThread::DeviceRenderThread(unsigned int index, IntersectionDevice *device,
		Scene *scn, const bool lowLatency) : RenderThread(index, scn) {
	intersectionDevice = device;

	// Allocate buffers

	// Sample buffer
	const size_t sampleBufferSize = lowLatency ? (SAMPLE_BUFFER_SIZE / 4) : SAMPLE_BUFFER_SIZE;
	sampleBuffer = new SampleBuffer(sampleBufferSize);

	// Ray buffer
	const size_t rayBufferSize = lowLatency ? (RAY_BUFFER_SIZE / 8) : RAY_BUFFER_SIZE;
	sampler = new RandomSampler(lowLatency, threadIndex + 1,
		scene->camera->film->GetWidth(), scene->camera->film->GetHeight());
	for(size_t i = 0; i < DEVICE_RENDER_BUFFER_COUNT; i++) {
		pathIntegrators[i] = new PathIntegrator(scene, sampler, sampleBuffer);
		rayBuffers[i] = new RayBuffer(rayBufferSize);
		rayBuffers[i]->PushUserData(i);
	}

	renderThread = NULL;
}

DeviceRenderThread::~DeviceRenderThread() {
	if (started)
		Stop();

	for(size_t i = 0; i < DEVICE_RENDER_BUFFER_COUNT; i++) {
		delete rayBuffers[i];
		delete pathIntegrators[i];
	}
	delete sampler;
	delete sampleBuffer;
}

void DeviceRenderThread::Start() {
	RenderThread::Start();

	sampler->Init(scene->camera->film->GetWidth(), scene->camera->film->GetHeight());
	sampleBuffer->Reset();
	for(size_t i = 0; i < DEVICE_RENDER_BUFFER_COUNT; i++) {
		rayBuffers[i]->Reset();
		rayBuffers[i]->ResetUserData();
		rayBuffers[i]->PushUserData(i);
		pathIntegrators[i]->ReInit();
	}

	intersectionDevice->Start();

	// Create the thread for the rendering
	renderThread = new boost::thread(boost::bind(DeviceRenderThread::RenderThreadImpl, this));
}

void DeviceRenderThread::Interrupt() {
	if (renderThread)
		renderThread->interrupt();
}

void DeviceRenderThread::Stop() {
	if (renderThread) {
		renderThread->interrupt();
		renderThread->join();
		delete renderThread;
		renderThread = NULL;
	}

	if (started)
		intersectionDevice->Stop();

	RenderThread::Stop();
}

void DeviceRenderThread::ClearPaths() {
	for(size_t i = 0; i < DEVICE_RENDER_BUFFER_COUNT; i++)
		pathIntegrators[i]->ClearPaths();
}

void DeviceRenderThread::RenderThreadImpl(DeviceRenderThread *renderThread) {
	cerr << "[DeviceRenderThread::" << renderThread->threadIndex << "] Rendering thread started" << endl;

	try {
		std::deque<RayBuffer *> todoBuffers;

		for(size_t i = 0; i < DEVICE_RENDER_BUFFER_COUNT; i++)
			todoBuffers.push_back(renderThread->rayBuffers[i]);

		while (!boost::this_thread::interruption_requested()) {
			// Produce buffers to trace
			while (todoBuffers.size() > 0) {
				RayBuffer *rayBuffer = todoBuffers.front();
				todoBuffers.pop_front();

				rayBuffer->Reset();
				renderThread->pathIntegrators[rayBuffer->GetUserData()]->FillRayBuffer(rayBuffer);
				renderThread->intersectionDevice->PushRayBuffer(rayBuffer);
			}

			RayBuffer *rayBuffer = renderThread->intersectionDevice->PopRayBuffer();
			renderThread->pathIntegrators[rayBuffer->GetUserData()]->AdvancePaths(rayBuffer);
			todoBuffers.push_back(rayBuffer);
		}

		cerr << "[DeviceRenderThread::" << renderThread->threadIndex << "] Rendering thread halted" << endl;
	} catch (boost::thread_interrupted) {
		cerr << "[DeviceRenderThread::" << renderThread->threadIndex << "] Rendering thread halted" << endl;
	} catch (cl::Error err) {
		cerr << "[DeviceRenderThread::" << renderThread->threadIndex << "] RenderingERROR: " << err.what() << "(" << err.err() << ")" << endl;
	}
}
